//------------------------------------------------------------
//
//    @@@@@\wvZ
//
//           < mech.cpp >
//
//
//      2002. 8.14 쐬 by Y
//      2003.10.13  by Y
//      2008. 1.12  by Y
//      2008. 1.14  by ╍
//      2010. 8.21  by ╍
//      2011.10.10  by ╍
//      2018. 4.15  by ╍
//
//------------------------------------------------------------

#include <stdafx.h>
#include <stdio.h>
#include <math.h>
#include "mech.h"
#include "matrix.h"

//-----------------------------------------------------------
//
//     ]߂̉^
//
//   j1       : ̑΋̈ʒuxNgƂ̔
//   l        : ߒ
//   theta    : ߊpψʂƂ̔
//   phi      : O̐߂̊pψʂƂ̔
//   j2       : [̑΋̈ʒuxNgƂ̔
//
//-----------------------------------------------------------

void crank_input( POINT j1, double l, ANGLE theta, ANGLE phi, POINT *j2 )
{
	double thephi, dthephi, ddthephi;
	double c,s;

	thephi = theta.the + phi.the;
	dthephi = theta.dthe + phi.dthe;
	ddthephi = theta.ddthe + phi.ddthe;

	c = cos( thephi );
	s = sin( thephi );

	j2->P.x   = j1.P.x + l*c;
	j2->DP.x  = j1.DP.x - l*s*dthephi;
	j2->DDP.x = j1.DDP.x - l*( s*ddthephi + c*dthephi*dthephi );

	j2->P.y   = j1.P.y + l*s;
	j2->DP.y  = j1.DP.y + l*c*dthephi;
	j2->DDP.y = j1.DDP.y + l*( c*ddthephi - s*dthephi*dthephi );
}
//-----------------------------------------------------------
//
//     i߂̉^
//
//   pg       : i΋̃KCḧʒuK肷邽߂
//              ̓_̈ʒuxNgƂ̔
//   psig     : i΋̃KCh̒̎ppƂ̔
//   s        : i΋͕̓ψʂƂ̔
//   gsi, eta : i΋̑΋_̓W
//   jp       : i΋̕ψʂƂ̔
//   jc       : i΋̂̑΋_̕ψʂƂ̔
//
//-----------------------------------------------------------

void slider_input( POINT pg, double l, ANGLE psig,  VARIABLE s, double gsi, double eta, POINT *jp, POINT *jc )
{
	double cp, sp;

	cp = cos( psig.the );
	sp = sin( psig.the );

	jp->P.x   = s.v*cp + pg.P.x;
	jp->DP.x  = s.dv*cp - s.v*sp*psig.dthe + pg.DP.x;
	jp->DDP.x = s.ddv*cp - 2.0*s.dv*sp*psig.dthe - s.v*sp*psig.ddthe - s.v*cp*psig.dthe*psig.dthe + pg.DDP.x;

	jp->P.y   = s.v*sp + pg.P.y;
	jp->DP.y  = s.dv*sp + s.v*cp*psig.dthe + pg.DP.y;
	jp->DDP.y = s.ddv*sp + 2.0*s.dv*cp*psig.dthe + s.v*cp*psig.ddthe - s.v*sp*psig.dthe*psig.dthe + pg.DDP.y;

	jc->P.x   =  gsi*cp - eta*sp + jp->P.x;
	jc->DP.x  = -gsi*sp*psig.dthe - eta*cp*psig.dthe + jp->DP.x;
	jc->DDP.x = -gsi*( sp*psig.ddthe + cp*psig.dthe*psig.dthe ) - eta*( cp*psig.ddthe - sp*psig.dthe*psig.dthe ) + jp->DDP.x;

	jc->P.y   =  gsi*sp + eta*cp + jp->P.y;
	jc->DP.y  =  gsi*cp*psig.dthe - eta*sp*psig.dthe + jp->DP.y;
	jc->DDP.y =  gsi*( cp*psig.ddthe - sp*psig.dthe*psig.dthe ) - eta*( sp*psig.ddthe + cp*psig.dthe*psig.dthe ) + jp->DDP.y;
}

//-----------------------------------------------------------
//
//     Q_ԂԐ̐p
//
//   j1, j2   : ʒuxNgƂ̔
//   *angle   : ̐p
//   ֐l   : 0 : I
//              1 : ُIiQ_vj
//
//-----------------------------------------------------------

int segment_angle( POINT j1, POINT j2, ANGLE *angle )
{
	POINT diff;
	double c, s;
	double absx, absy;
	double q1, q2;
	double r, dr;

	get_diff( j1, j2, &diff );
	absx = fabs( diff.P.x );
	absy = fabs( diff.P.y );

	if( absx < 10e-20 && absy < 10e-20 ){
		angle->the   = 0.0;
		angle->dthe  = 0.0;
		angle->ddthe = 0.0;
		printf("++ Error in segment_angle \n");
		return ERROR;
	}
	angle->the = atan2( diff.P.y, diff.P.x );
	c = cos( angle->the );
	s = sin( angle->the );

	if( absy > 10e-20 && absx/absy < 10e-10 ){
		r   = sqrt( diff.P.x*diff.P.x + diff.P.y*diff.P.y );
		dr  = ( diff.DP.x*diff.P.x + diff.DP.y*diff.P.y )/r;
		
		if( s >= 0.0 ){
			angle->dthe  = -diff.DP.x/r;
			angle->ddthe = ( -diff.DDP.x - 2.0*dr*angle->dthe )/r;
			return SUCCESS;
		}
		else{
			angle->dthe  = diff.DP.x/r;
			angle->ddthe = ( diff.DDP.x - 2.0*dr*angle->dthe )/r;
			return SUCCESS;
		}
	}
	else{
		q1 = diff.DP.y*diff.P.x - diff.P.y*diff.DP.x;
		q2 = diff.DDP.y*diff.P.x - diff.P.y*diff.DDP.x;
		angle->dthe = q1*c*c/(diff.P.x*diff.P.x);
		angle->ddthe = ( (diff.P.x*q2 - 2.0*diff.DP.x*q1)*c - 2.0*diff.P.x*angle->dthe*q1*s)*c/( diff.P.x*diff.P.x*diff.P.x );
		return SUCCESS;
	}
}

//-----------------------------------------------------------
//
//       ]΋ŐڑꂽQA߂̉^
//
//   j1, j2   : QAߗ[̑΋̈ʒuxNgƂ̔
//   l1, l2   : ߒ
//   minv     : @\̔]̎wCWP[^
//   *jb      : ̑΋̈ʒuxNgƂ̔
//   ֐l   : 0 : I
//              1 : ُIiA\łj
//
//-----------------------------------------------------------

int RRR_links( POINT j1, POINT j2, double l1, double l2, int minv, POINT *jb )
{
	POINT diff;
	double q2;
	double q, dq, ddq;

	double cb, sb, ss;
	ANGLE angle;
	double phi, dphi, ddphi;
	double bet, dbet, ddbet;
	double cp, sp;
	int ier;

	get_diff( j1, j2, &diff );
	q2 = diff.P.x*diff.P.x + diff.P.y*diff.P.y;

	if( q2 <= 10e-30 ){
		printf("++ Error in RRR_links (Q΋􂪈vĂ܂.) \n");
		return ERROR;
	}

	q = sqrt( q2 );
	dq = ( diff.DP.x*diff.P.x + diff.DP.y*diff.P.y )/q;
	ddq = ( ( diff.DDP.x*diff.P.x + diff.DP.x*diff.DP.x
		+ diff.DDP.y*diff.P.y + diff.DP.y*diff.DP.y )*q
		- ( diff.DP.x*diff.P.x + diff.DP.y*diff.P.y )*dq )/q2;

	cb = ( l1*l1 + q2 - l2*l2 )/( 2.0*l1*q );
	if( fabs( cb ) > 1.0 ){
		printf("++ Error in RRR_links (A\ł܂D)\n");
		return ERROR;
	}
	sb = sqrt( 1.0 - cb*cb );
	ss = l1*l1 - l2*l2 - q2;

	ier = segment_angle( j1, j2, &angle );
	if( ier != SUCCESS ){
  		printf("++ Error in RRR_links (Q΋􂪈vĂ܂.) \n");
		return ERROR;
	}

	bet = acos( cb );
	dbet = ss*dq/( 2.0*l1*q2*sb );
	ddbet = ( ( ss*( ddq*q - 2.0*dq*dq ) - 2.0*( dq*q )*( dq*q) )*sb - ss*dq*q*dbet*cb )/( 2.0*l1*q2*q*sb*sb );

	if( minv > 0 ) {
		phi = angle.the + bet;
		dphi = angle.dthe + dbet;
		ddphi = angle.ddthe + ddbet;
	}
	else {
		phi = angle.the - bet;
		dphi = angle.dthe - dbet;
		ddphi = angle.ddthe - ddbet;
	}

	cp = cos( phi );
	sp = sin( phi );

	jb->P.x = j1.P.x + l1*cp;
	jb->P.y = j1.P.y + l1*sp;
	jb->DP.x = j1.DP.x - l1*dphi*sp;
	jb->DP.y = j1.DP.y + l1*dphi*cp;
	jb->DDP.x = j1.DDP.x - l1*( ddphi*sp + dphi*dphi*cp );
	jb->DDP.y = j1.DDP.y + l1*( ddphi*cp - dphi*dphi*sp );

	return SUCCESS;
}
//-----------------------------------------------------------
//
//       i΋ƂQ̉]΋ŐڑꂽQA߂̉^
//
//   jr       : []΋̈ʒuxNgƂ̔
//   l        : Q]΋Ԃ̋iߒj
//   pg       : i΋̃KCḧʒuK肷邽߂
//              ̓_̈ʒuxNgƂ̔
//   psig     : i΋̃KCh̒̎ppƂ̔
//   gsi, eta : i΋̑΋_̓W
//   minv     : @\̔]̎wCWP[^
//   *s       : i΋̕ψʂƂ̔
//   *jp      : i΋̕ψʂƂ̔
//   *jb      : ̉]΋̈ʒuxNgƂ̔
//   ֐l   : 0 : I
//              1 : ُIiA\łj
//
//-----------------------------------------------------------

int PRR_links( POINT jr, double l, POINT pg, ANGLE psig, double gsi, double eta, int minv, VARIABLE *s, POINT *jp, POINT *jb )
{
	double cp, sp;
	double a, da, dda, b, db, ddb, d, dd, ddd, q, dq, ddq, s0, ds0, dds0;

	cp = cos( psig.the );
	sp = sin( psig.the );

	a = -eta*sp + pg.P.x - jr.P.x;
	b =  eta*cp + pg.P.y - jr.P.y;

	d = l*l - ( a*sp - b*cp )*( a*sp - b*cp );
	if( d<0.0 )  return ERROR;

	da = -eta*cp*psig.dthe + pg.DP.x - jr.DP.x;
	dda = -eta*( cp*psig.ddthe - sp*psig.dthe*psig.dthe ) + pg.DDP.x - jr.DDP.x;

	db = -eta*sp*psig.dthe + pg.DP.y - jr.DP.y;
	ddb = -eta*( sp*psig.ddthe + cp*psig.dthe*psig.dthe ) + pg.DDP.y - jr.DDP.y;

	dd = -2.0*( da*sp + a*cp*psig.dthe - db*cp + b*sp*psig.dthe )*( a*sp - b*cp );
	ddd = -2.0*( ( dda*sp + 2.0*da*cp*psig.dthe + a*cp*psig.ddthe - a*sp*psig.dthe*psig.dthe - ddb*cp + 2.0*db*sp*psig.dthe + b*sp*psig.ddthe + b*cp*psig.dthe*psig.dthe )*( a*sp - b*cp ) + ( da*sp + a*cp*psig.dthe - db*cp + b*sp*psig.dthe )*( da*sp + a*cp*psig.dthe - db*cp + b*sp*psig.dthe ) );

	q = sqrt( d );
	dq = dd/( 2.0*q );
	ddq = ( 2.0*ddd*d - dd*dd )/( 4.0*d*q );

    s0 = -a*cp - b*sp - gsi;
    ds0 = -da*cp + a*sp*psig.dthe - db*sp - b*cp*psig.dthe;
	dds0 = -dda*cp + 2.0*da*sp*psig.dthe + a*sp*psig.ddthe + a*cp*psig.dthe*psig.dthe - ddb*sp - 2.0*db*cp*psig.dthe - b*cp*psig.ddthe + b*sp*psig.dthe*psig.dthe;

	if( minv>0 ) {
		s->v = s0 + q;
		s->dv = ds0 + dq;
		s->ddv = dds0 + ddq;
	}
	else {
		s->v = s0 - q;
		s->dv = ds0 - dq;
		s->ddv = dds0 - ddq;
	}

	jp->P.x = s->v*cp + pg.P.x;
	jp->P.y = s->v*sp + pg.P.y;

	jp->DP.x = s->dv*cp - s->v*sp*psig.dthe + pg.DP.x;
	jp->DP.y = s->dv*sp + s->v*cp*psig.dthe + pg.DP.y;
		
	jp->DDP.x = s->ddv*cp - 2.0*s->dv*sp*psig.dthe - s->v*sp*psig.ddthe - s->v*cp*psig.dthe*psig.dthe + pg.DDP.x;
	jp->DDP.y = s->ddv*sp + 2.0*s->dv*cp*psig.dthe + s->v*cp*psig.ddthe - s->v*sp*psig.dthe*psig.dthe + pg.DDP.y;

	jb->P.x = gsi*cp - eta*sp + jp->P.x;
	jb->P.y = gsi*sp + eta*cp + jp->P.y;

	jb->DP.x = -gsi*sp*psig.dthe - eta*cp*psig.dthe + jp->DP.x;
	jb->DP.y =  gsi*cp*psig.dthe - eta*sp*psig.dthe + jp->DP.y;

	jb->DDP.x = -gsi*( sp*psig.ddthe + cp*psig.dthe*psig.dthe ) - eta*( cp*psig.ddthe - sp*psig.dthe*psig.dthe ) + jp->DDP.x;
	jb->DDP.y =  gsi*( cp*psig.ddthe - sp*psig.dthe*psig.dthe ) - eta*( sp*psig.ddthe + cp*psig.dthe*psig.dthe ) + jp->DDP.y;

	return SUCCESS;
}
//-----------------------------------------------------------
//
//     ߂̊pψʁiߏ̂Q΋ԂԐ̐pj
//
//   j1, j2   : ʒuxNgƂ̔
//   *angle   : ̐p
//   ֐l   : 0 : I
//              1 : ُIiQ_vj
//
//-----------------------------------------------------------

int link_angle( POINT j1, POINT j2, ANGLE *angle )
{
	POINT diff;
	double q;

	get_diff( j1, j2, &diff );

	q = diff.P.x*diff.P.x + diff.P.y*diff.P.y;
	if( q <= 10e-30 ){
		printf("++ Error in link_angle iQ΋􂪈vĂ܂j\n");
		return ERROR;
	}

	angle->the   = atan2( diff.P.y, diff.P.x );
	angle->dthe  = ( diff.DP.y*diff.P.x - diff.P.y*diff.DP.x )/q;
	angle->ddthe = ( diff.DDP.y*diff.P.x - diff.P.y*diff.DDP.x )/q;

	return SUCCESS;
}

//-----------------------------------------------------------
//
//     Ԑߏ̕`_̉^
//
//   j1, j2   : ʒuxNgƂ̔
//   gsi, eta : `_̒Ԑߏ̓W
//   *jc      : `_̐Î~Wn̈ʒuxNgƂ̔
//   *angle   : Ԑ߂̐pƂ̔
//   ֐l   : 0 : I
//              1 : ُIiQ_vj
//
//-----------------------------------------------------------

int coupler_point( POINT j1, POINT j2, double gsi, double eta, POINT *jc, ANGLE *angle )
{
	double c, s;
	double q1, q2;
	int ier;

	ier = link_angle( j1, j2, angle );
	if( ier != SUCCESS ) {
		printf("++ Error in coupler_point iQ΋􂪈vĂ܂j\n");
		return ERROR;
	}

	c = cos( angle->the );
	s = sin( angle->the );

	q1 = gsi*c - eta*s;
	q2 = gsi*s + eta*c;

	jc->P.x = j1.P.x + q1;
	jc->P.y = j1.P.y + q2;

	jc->DP.x = j1.DP.x - q2*angle->dthe;
	jc->DP.y = j1.DP.y + q1*angle->dthe;

	jc->DDP.x = j1.DDP.x - q2*angle->ddthe - q1*angle->dthe*angle->dthe;
	jc->DDP.y = j1.DDP.y + q1*angle->ddthe - q2*angle->dthe*angle->dthe;

	return SUCCESS;
}

//-----------------------------------------------------------
//
//     [pɂWϊs
//
//   Roll_Angle 
//            : [p
//   TT[3][3] : Wϊs
//
//-----------------------------------------------------------

void RollTT ( double Roll_angle, double TT[3][3] )
{
	double CR, SR;

	CR = cos( Roll_angle );
	SR = sin( Roll_angle );

	TT[0][0] = 1.0;
	TT[1][0] = 0.0;
	TT[2][0] = 0.0;
	TT[0][1] = 0.0;
	TT[1][1] = CR;
	TT[2][1] = SR;
	TT[0][2] = 0.0;
	TT[1][2] = -SR;
	TT[2][2] = CR;
}

//-----------------------------------------------------------
//
//     sb`pɂWϊs
//
//   Pitch_Angle 
//            : sb`p
//   TT[3][3] : Wϊs
//
//-----------------------------------------------------------

void PitchTT ( double Pitch_angle, double TT[3][3] )
{
	double CP, SP;

	CP = cos( Pitch_angle );
	SP = sin( Pitch_angle );

	TT[0][0] = CP;
	TT[1][0] = 0.0;
	TT[2][0] = -SP;
	TT[0][1] = 0.0;
	TT[1][1] = 1.0;
	TT[2][1] = 0.0;
	TT[0][2] = SP;
	TT[1][2] = 0.0;
	TT[2][2] = CP;
}

//-----------------------------------------------------------
//
//     [pɂWϊs
//
//   Yaw_Angle 
//            : [p
//   TT[3][3] : Wϊs
//
//-----------------------------------------------------------

void YawTT ( double Yaw_angle, double TT[3][3] )
{
	double CY, SY;

	CY = cos( Yaw_angle );
	SY = sin( Yaw_angle );

	TT[0][0] = CY;
	TT[1][0] = SY;
	TT[2][0] = 0.0;
	TT[0][1] = -SY;
	TT[1][1] = CY;
	TT[2][1] = 0.0;
	TT[0][2] = 0.0;
	TT[1][2] = 0.0;
	TT[2][2] = 1.0;
}

//-----------------------------------------------------------
//
//     [Esb`E[pWϊs߂
//
//   RPY_ANGLES 
//            : [Esb`E[pƂ̔
//   TT[3][3], dTT[3][3], ddTT[3][3] 
//            : WϊsƂ̔
//
//-----------------------------------------------------------

void RPYTT ( POSTURE RPY_ANGLES, double TT[3][3], double dTT[3][3], double ddTT[3][3] )
{
	double CR, SR, CP, SP, CY, SY;
	int i, j;
	double WA1[3][3], WA2[3][3], WA3[3][3], 
		   WB1[3][3], WB2[3][3], WB3[3][3], 
		   WC1[3][3], WC2[3][3], WC3[3][3];

	CR = cos( RPY_ANGLES.RPY.x );
	SR = sin( RPY_ANGLES.RPY.x );
	CP = cos( RPY_ANGLES.RPY.y );
	SP = sin( RPY_ANGLES.RPY.y );
	CY = cos( RPY_ANGLES.RPY.z );
	SY = sin( RPY_ANGLES.RPY.z );

	WA1[0][0] = 0.0;
	WA1[1][0] = -SR*SY + CR*SP*CY;
	WA1[2][0] = CR*SY + SR*SP*CY;
	WA1[0][1] = 0.0;
	WA1[1][1] = -SR*CY - CR*SP*SY;
	WA1[2][1] = CR*CY - SR*SP*SY;
	WA1[0][2] = 0.0;
	WA1[1][2] = -CR*CP;
	WA1[2][2] = -SR*CP;

	WA2[0][0] = -SP*CY;
	WA2[1][0] =  SR*CP*CY;
	WA2[2][0] = -CR*CP*CY;
	WA2[0][1] =  SP*SY;
	WA2[1][1] = -SR*CP*SY;
	WA2[2][1] =  CR*CP*SY;
	WA2[0][2] =  CP;
	WA2[1][2] =  SR*SP;
	WA2[2][2] = -CR*SP;

	WA3[0][0] = -CP*SY;
	WA3[1][0] =  CR*CY - SR*SP*SY;
	WA3[2][0] =  SR*CY + CR*SP*SY;
	WA3[0][1] = -CP*CY;
	WA3[1][1] = -CR*SY - SR*SP*CY;
	WA3[2][1] = -SR*SY + CR*SP*CY;
	WA3[0][2] =  0.0;
	WA3[1][2] =  0.0;
	WA3[2][2] =  0.0;

	WB1[0][0] =  0.0;
	WB1[1][0] = -CR*SY - SR*SP*CY;
	WB1[2][0] = -SR*SY + CR*SP*CY;
	WB1[0][1] =  0.0;
	WB1[1][1] = -CR*CY + SR*SP*SY;
	WB1[2][1] = -SR*CY - CR*SP*SY;
	WB1[0][2] =  0.0;
	WB1[1][2] =  SR*CP;
	WB1[2][2] = -CR*CP;

	WB2[0][0] = -CP*CY;
	WB2[1][0] = -SR*SP*CY;
	WB2[2][0] =  CR*SP*CY;
	WB2[0][1] =  CP*SY;
	WB2[1][1] =  SR*SP*SY;
	WB2[2][1] = -CR*SP*SY;
	WB2[0][2] =  -SP;
	WB2[1][2] =  SR*CP;
	WB2[2][2] = -CR*CP;

	WB3[0][0] = -CP*CY;
	WB3[1][0] = -CR*SY - SR*SP*CY;
	WB3[2][0] = -SR*SY + CR*SP*CY;
	WB3[0][1] =  CP*SY;
	WB3[1][1] = -CR*CY + SR*SP*SY;
	WB3[2][1] = -SR*CY - CR*SP*SY;
	WB3[0][2] = 0.0;
	WB3[1][2] = 0.0;
	WB3[2][2] = 0.0;

	WC1[0][0] = 0.0;
	WC1[1][0] =  CR*CP*CY;
	WC1[2][0] =  SR*CP*CY;
	WC1[0][1] = 0.0;
	WC1[1][1] = -CR*CP*SY;
	WC1[2][1] = -SR*CP*SY;
	WC1[0][2] = 0.0;
	WC1[1][2] =  CR*SP;
	WC1[2][2]=   SR*SP;

	WC2[0][0] =  SP*SY;
	WC2[1][0] = -SR*CP*SY;
	WC2[2][0] =  CR*CP*SY;
	WC2[0][1] =  SP*CY;
	WC2[1][1] = -SR*CP*CY;
	WC2[2][1] =  CR*CP*CY;
	WC2[0][2] = 0.0;
	WC2[1][2] = 0.0;
	WC2[2][2] = 0.0;

	WC3[0][0] = 0.0;
	WC3[1][0] = -SR*CY - CR*SP*SY;
	WC3[2][0] =  CR*CY - SR*SP*SY;
	WC3[0][1] = 0.0;
	WC3[1][1] =  SR*SY - CR*SP*CY;
	WC3[2][1] = -CR*SY - SR*SP*CY;
	WC3[0][2] = 0.0;
	WC3[1][2] = 0.0;
	WC3[2][2] = 0.0;

	TT[0][0] =  CP*CY;
	TT[1][0] =  CR*SY + SR*SP*CY;
	TT[2][0] =  SR*SY - CR*SP*CY;
	TT[0][1] = -CP*SY;
	TT[1][1] =  CR*CY - SR*SP*SY;
	TT[2][1] =  SR*CY + CR*SP*SY;
	TT[0][2] =  SP;
	TT[1][2] = -SR*CP;
	TT[2][2] =  CR*CP;

	for(i = 0; i < 3; i++){
		for(j = 0; j < 3; j++){
			dTT[i][j]  = WA1[i][j]*RPY_ANGLES.DRPY.x
				       + WA2[i][j]*RPY_ANGLES.DRPY.y
					   + WA3[i][j]*RPY_ANGLES.DRPY.z;
			ddTT[i][j] = WA1[i][j]*RPY_ANGLES.DDRPY.x 
				       + WA2[i][j]*RPY_ANGLES.DDRPY.y
					   + WA3[i][j]*RPY_ANGLES.DDRPY.z
					   + WB1[i][j]*RPY_ANGLES.DRPY.x*RPY_ANGLES.DRPY.x
					   + WB2[i][j]*RPY_ANGLES.DRPY.y*RPY_ANGLES.DRPY.y
					   + WB3[i][j]*RPY_ANGLES.DRPY.z*RPY_ANGLES.DRPY.z
					   + 2.0*WC1[i][j]*RPY_ANGLES.DRPY.x*RPY_ANGLES.DRPY.y
					   + 2.0*WC2[i][j]*RPY_ANGLES.DRPY.y*RPY_ANGLES.DRPY.z
					   + 2.0*WC3[i][j]*RPY_ANGLES.DRPY.z*RPY_ANGLES.DRPY.x;
		}
	}
}

//-----------------------------------------------------------
//
//     Wϊs񂩂烍[Esb`E[p߂
//
//   TT[3][3], dTT[3][3], ddTT[3][3] 
//            : WϊsƂ̔
//   *RPY_ANGLES 
//            : [Esb`E[pƂ̔
//
//-----------------------------------------------------------

void TTRPY ( double TT[3][3], double dTT[3][3], double ddTT[3][3], POSTURE *RPY_ANGLES  )
{
	double CP, SP, CY, SY, QR, QY;

	RPY_ANGLES->RPY.y = asin( TT[0][2] );
	CP = cos( RPY_ANGLES->RPY.y );
	SP = sin( RPY_ANGLES->RPY.y );

	if( fabs( CP )>=1.0e-5 )  {
		RPY_ANGLES->RPY.x = atan2( -TT[1][2]/CP , TT[2][2]/CP );
		RPY_ANGLES->RPY.z = atan2( -TT[0][1]/CP , TT[0][0]/CP );

		RPY_ANGLES->DRPY.y = dTT[0][2]/CP;
		RPY_ANGLES->DDRPY.y = ( ddTT[0][2]*CP + dTT[0][2]*RPY_ANGLES->DRPY.y*SP )/CP/CP;

			QR = TT[2][2]*TT[2][2] + TT[1][2]*TT[1][2];
		RPY_ANGLES->DRPY.x = ( dTT[2][2]*TT[1][2] - dTT[1][2]*TT[2][2] )/QR;
		RPY_ANGLES->DDRPY.x = ( ddTT[2][2]*TT[1][2]*QR
							- 2.0*dTT[2][2]*dTT[2][2]*TT[2][2]*TT[1][2]
							+ 2.0*dTT[2][2]*dTT[1][2]
								*( TT[2][2]*TT[2][2] - TT[1][2]*TT[1][2] )
							- ddTT[1][2]*TT[2][2]*QR
							+ 2.0*dTT[1][2]*dTT[1][2]*TT[2][2]*TT[1][2] )/QR/QR;

			QY = TT[0][0]*TT[0][0] + TT[0][1]*TT[0][1];
		RPY_ANGLES->DRPY.z = ( dTT[0][0]*TT[0][1] - dTT[0][1]*TT[0][0] )/QY;
		RPY_ANGLES->DDRPY.z = ( ddTT[0][0]*TT[0][1]*QY
							- 2.0*dTT[0][0]*dTT[0][0]*TT[0][0]*TT[0][1]
							+ 2.0*dTT[0][0]*dTT[0][1]
								*( TT[0][0]*TT[0][0] - TT[0][1]*TT[0][1] )
							- ddTT[0][1]*TT[0][0]*QY
							+ 2.0*dTT[0][1]*dTT[0][1]*TT[0][0]*TT[0][1] )/QY/QY;
	}
	else {
		RPY_ANGLES->RPY.x = 0.0;
		RPY_ANGLES->RPY.z = atan2( TT[1][0] , TT[1][1] );
			CY = cos( RPY_ANGLES->RPY.z );
			SY = sin( RPY_ANGLES->RPY.z );
		RPY_ANGLES->DRPY.x = 0.0;
		RPY_ANGLES->DDRPY.x = 0.0;

		if( fabs( CY ) >= 1.0e-5 )  {
			RPY_ANGLES->DRPY.y = -dTT[0][0]/SP/CY;
			RPY_ANGLES->DRPY.z = dTT[1][0]/CY;
			RPY_ANGLES->DDRPY.y = ( -ddTT[0][0] 
					+ 2.0*RPY_ANGLES->DRPY.y*RPY_ANGLES->DRPY.z*SY*SP )/SP/CY;
			RPY_ANGLES->DDRPY.z = ( ddTT[1][0] 
					+ RPY_ANGLES->DRPY.z*RPY_ANGLES->DRPY.z*SY )/CY;
		}
		else {
			RPY_ANGLES->DRPY.y = dTT[0][1]/SP/SY;
			RPY_ANGLES->DRPY.z = -dTT[1][1]/SY;
			RPY_ANGLES->DDRPY.y = ddTT[0][1]/SY/SP;
			RPY_ANGLES->DDRPY.z = -ddTT[1][1]/SY;
		}
	}
}

//-----------------------------------------------------------
//
//     [Esb`E[ppxEpxxNg߂
//
//   RPY_ANGLES 
//            : [Esb`E[pƂ̔
//   AVELO    : pxƊpx
//
//-----------------------------------------------------------

void RPYAVELO ( POSTURE RPY_ANGLES, AVELOCITY *AVELO )
{
	double CR, SR, CP, SP, CY, SY;
	double DARP, DAPY, DAYR;

	CR = cos( RPY_ANGLES.RPY.x );
	SR = sin( RPY_ANGLES.RPY.x );
	CP = cos( RPY_ANGLES.RPY.y );
	SP = sin( RPY_ANGLES.RPY.y );
	CY = cos( RPY_ANGLES.RPY.z );
	SY = sin( RPY_ANGLES.RPY.z );

	DARP = RPY_ANGLES.DRPY.x*RPY_ANGLES.DRPY.y;
	DAPY = RPY_ANGLES.DRPY.y*RPY_ANGLES.DRPY.z;
	DAYR = RPY_ANGLES.DRPY.z*RPY_ANGLES.DRPY.x;

	AVELO->OMG.x =  RPY_ANGLES.DRPY.x*CY*CP + RPY_ANGLES.DRPY.y*SY;
	AVELO->OMG.y = -RPY_ANGLES.DRPY.x*SY*CP + RPY_ANGLES.DRPY.y*CY;
	AVELO->OMG.z =  RPY_ANGLES.DRPY.x*SP    + RPY_ANGLES.DRPY.z;

	AVELO->DOMG.x =  RPY_ANGLES.DDRPY.x*CY*CP + RPY_ANGLES.DDRPY.y*SY + DAPY*CY - DARP*CY*SP - DAYR*SY*CP;
	AVELO->DOMG.y = -RPY_ANGLES.DDRPY.x*SY*CP + RPY_ANGLES.DDRPY.y*CY - DAPY*SY + DARP*SY*SP - DAYR*CY*CP;
    AVELO->DOMG.z =  RPY_ANGLES.DDRPY.x*SP    + RPY_ANGLES.DDRPY.z    + DARP*CP;
}

//----------------------------------------------------
//
//     WϊsɊÂRWϊ
//
//   PM @    : Wn̈ʒuxNg
//   P0       : Wn_̐Î~n̈ʒuxNg
//              Ƃ̔
//   TT[3][3], dTT[3][3], ddTT[3][3] 
//            : WϊsƂ̔
//   PF       : Î~n̈ʒuxNgƂ̔
//
//---------------------------------------------------
void TRANS( VECTOR3 PM, POSITION P0,
            double TT[3][3], double dTT[3][3], double ddTT[3][3],
			POSITION *PF )
{
	MPV3( TT, PM, &PF->P  );
	MPV3( dTT, PM, &PF->DP );
	MPV3( ddTT, PM, &PF->DDP );

	PF->P.x = PF->P.x + P0.P.x;
	PF->P.y = PF->P.y + P0.P.y;
	PF->P.z = PF->P.z + P0.P.z;

	PF->DP.x = PF->DP.x + P0.DP.x;
	PF->DP.y = PF->DP.y + P0.DP.y;
	PF->DP.z = PF->DP.z + P0.DP.z;

	PF->DDP.x = PF->DDP.x + P0.DDP.x;
	PF->DDP.y = PF->DDP.y + P0.DDP.y;
	PF->DDP.z = PF->DDP.z + P0.DDP.z;
}

//-----------------------------------------------------------
//
//     ԉ]߂̉^
//
//   j1       : ̑΋_̈ʒuxNgƂ̔
//   TT1[3][3], dTT1[3][3], ddTT1[3][3]
//            : ɌŒ肵Wn̍WϊsƂ̔
//   theta    : ߂̋쓮]pψʂƂ̔
//   j2m      : ɌŒ肵Wnɂ[̓Wn_
//              ̈ʒuxNg
//   TT12[3][3]
//            : ɌŒ肵Wn[ɌŒ肵Wn
//              ւ̍Wϊs
//   j2       : [̑΋_̈ʒuxNgƂ̔
//   TT2[3][3], dTT2[3][3], ddTT2[3][3]
//            : [ɌŒ肵Wn̍WϊsƂ̔
//
//-----------------------------------------------------------

void s_crank_input( POSITION j1, 
			double TT1[3][3], double dTT1[3][3], double ddTT1[3][3],
			ANGLE theta,
			VECTOR3 j2m,
			double TT12[3][3],
			POSITION *j2, 
			double TT2[3][3], double dTT2[3][3], double ddTT2[3][3] )
{
	double W1[3][3], W2[3][3], W3[3][3], W4[3][3], W5[3][3], W6[3][3];
	double CT, ST, DCT, DST, DDCT, DDST;
//
	CT = cos( theta.the );
	ST = sin( theta.the );
	DCT = -theta.dthe*ST;
	DST = theta.dthe*CT;
	DDCT = -theta.ddthe*ST - theta.dthe*theta.dthe*CT;
	DDST = theta.ddthe*CT - theta.dthe*theta.dthe*ST;
//
	W1[0][0] = CT;
	W1[0][1] = -ST;
	W1[0][2] = 0.0;
	W1[1][0] = ST;
	W1[1][1] = CT;
	W1[1][2] = 0.0;
	W1[2][0] = 0.0;
	W1[2][1] = 0.0;
	W1[2][2] = 1.0;
//
	W2[0][0] = DCT;
	W2[0][1] = -DST;
	W2[0][2] = 0.0;
	W2[1][0] = DST;
	W2[1][1] = DCT;
	W2[1][2] = 0.0;
	W2[2][0] = 0.0;
	W2[2][1] = 0.0;
	W2[2][2] = 0.0;
//
	W3[0][0] = DDCT;
	W3[0][1] = -DDST;
	W3[0][2] = 0.0;
	W3[1][0] = DDST;
	W3[1][1] = DDCT;
	W3[1][2] = 0.0;
	W3[2][0] = 0.0;
	W3[2][1] = 0.0;
	W3[2][2] = 0.0;
//
	MPM3 ( TT1, W1,  W4 );
	MPV3 ( W4, j2m,  &j2->P );      
	VAV3 ( j2->P, j1.P, &j2->P );
//
	MPM3 ( dTT1, W1,  W4 );
	MPM3 ( TT1, W2,  W5 );
	MAM3 ( W4, W5, W4 );
	MPV3 ( W4, j2m,  &j2->DP );      
	VAV3 ( j2->DP, j1.DP, &j2->DP );
//
	MPM3 ( ddTT1, W1,  W4 );
	MPM3 ( dTT1, W2,  W5 );
	MPM3 ( TT1, W3,  W6 );
	MAM3 ( W4, W5,  W4 );
	MAM3 ( W4, W5,  W4 );
	MAM3 ( W4, W6,  W4 );
	MPV3 ( W4, j2m,  &j2->DDP );      
	VAV3 ( j2->DDP, j1.DDP,  &j2->DDP );
//
	MPM3 ( TT1, W1,  W4 );
	MPM3 ( W4, TT12,  TT2 );
//
	MPM3 ( dTT1, W1,  W4 );
	MPM3 ( TT1, W2,  W5 );
	MAM3 ( W4, W5,  W4 );
	MPM3 ( W4, TT12,  dTT2 );
//
	MPM3 ( ddTT1, W1,  W4 );
	MPM3 ( dTT1, W2,  W5 );
	MPM3 ( TT1, W3,  W6 );
	MAM3 ( W4, W5, W4 );
	MAM3 ( W4, W5, W4 );
	MAM3 ( W4, W6, W4 );
	MPM3 ( W4, TT12,  ddTT2 );
//
}

//-----------------------------------------------------------
//
//      |i|΋QA߂̉^
//
//   s1, s2   : [̋΋̈ʒuxNgƂ̔
//   p0       : s1p܂ł̌Œ苗
//   p        : i΋̈ʒuxNgƂ̔
//   the      : i΋ψʂƂ̔
//
//-----------------------------------------------------------

int s_SPS_links ( POSITION s1, POSITION s2,
				   double p0,
				   POSITION *p,
				   VARIABLE *the )
{
	double X, Y, Z, dX, dY, dZ, ddX, ddY, ddZ, TH, dTH, ddTH, w1, w2, w3;

	X = s2.P.x - s1.P.x;
	Y = s2.P.y - s1.P.y;
	Z = s2.P.z - s1.P.z;
	dX = s2.DP.x - s1.DP.x;
	dY = s2.DP.y - s1.DP.y;
	dZ = s2.DP.z - s1.DP.z;
	ddX = s2.DDP.x - s1.DDP.x;
	ddY = s2.DDP.y - s1.DDP.y;
	ddZ = s2.DDP.z - s1.DDP.z;

	TH = sqrt( X*X + Y*Y + Z*Z );
	if( fabs( TH ) < 1.0e-30 )  {
		printf( "++ Error in s_SPS_links (ُԂł)\n");
		return ERROR;
	}
	dTH = ( dX*X + dY*Y + dZ*Z )/TH;
	ddTH = ( ( ddX*X + dX*dX + ddY*Y + dY*dY + ddZ*Z + dZ*dZ )*TH - ( dX*X + dY*Y + dZ*Z )*dTH )/( TH*TH );

	w1 = p0/TH;
	w2 = -p0*dTH/( TH*TH );
	w3 = -p0*( ddTH  - 2.0*dTH*dTH*TH )/( TH*TH*TH*TH );
	
	p->P.x = w1*X + s1.P.x;
	p->P.y = w1*Y + s1.P.y;
	p->P.z = w1*Z + s1.P.z;

	p->DP.x = w2*X + w1*dX + s1.DP.x;
	p->DP.y = w2*Y + w1*dY + s1.DP.y;
	p->DP.z = w2*Z + w1*dZ + s1.DP.z;

	p->DDP.x = w3*X + 2.0*w2*dX + w1*ddX + s1.DDP.x;
	p->DDP.y = w3*Y + 2.0*w2*dY + w1*ddY + s1.DDP.y;
	p->DDP.z = w3*Z + 2.0*w2*dZ + w1*ddZ + s1.DDP.z;

	the->v = TH - p0;
	the->dv = dTH;
	the->ddv = ddTH;
	
	return SUCCESS;

}

//-----------------------------------------------------------
//
//      ||i΋QA߂̉^
//
//   s1  @@ : [̋΋̈ʒuxNgƂ̔
//   rs       : Q΋Ԃ̋
//   g        : i΋̒KCh\Ԓ̂P_
//              ʒuxNgƂ̔
//   TLT[3][3], dTLT[3][3], ddTLT[3][3] 
//            : g3_Ci΋̒KCh\Ԓ
//              ZƂ铮Wn̍WϊsƂ̔i]uj
//   s2p      : ̋΋̈ʒuxNg
//            ii΋ɌŒ肵Wnj
//   minv     : @\̔]̎wCWP[^
//   the      : i΋̕ψʂƂ̔
//   p        : i΋̈ʒuxNgƂ̔
//   s2  @@ : ̋΋̈ʒuxNgƂ̔
//
//-----------------------------------------------------------

int s_SSP_links ( POSITION s1,
                   double rs,
				   POSITION g,
                   double TLT[3][3], double dTLT[3][3], double ddTLT[3][3], 
                   VECTOR3 s2p,
				   int minv,
				   VARIABLE *the,
				   POSITION *p,
				   POSITION *s2 )
{
	VECTOR3 UVLF, PLS1F, dUVLF, dPLS1F, ddUVLF, ddPLS1F, ww, dww, ddww;
	double a, da, dda, b, db, ddb, d, rootd, w1, w2, w3, w4, w5;
	
	UVLF.x = TLT[0][2];
	UVLF.y = TLT[1][2];
	UVLF.z = TLT[2][2];
	PLS1F.x = g.P.x - s1.P.x;
	PLS1F.y = g.P.y - s1.P.y;
	PLS1F.z = g.P.z - s1.P.z;

	VIV3 ( UVLF, PLS1F, &a );
	a = a + s2p.z;

	MPV3 ( TLT, s2p, &ww );
	VIV3 ( ww, PLS1F, &b );
    b = b*2.0 + PLS1F.x*PLS1F.x + PLS1F.y*PLS1F.y + PLS1F.z*PLS1F.z + s2p.x*s2p.x + s2p.y*s2p.y + s2p.z*s2p.z - rs*rs;

	d = a*a - b;
	if( d<0 ) {
		printf("++ Error in s_SSP_links (A\ł܂.) \n");
		return ERROR;
	}
	rootd = sqrt( d );
	dUVLF.x = dTLT[0][2];
	dUVLF.y = dTLT[1][2];
	dUVLF.z = dTLT[2][2];
	dPLS1F.x = g.DP.x - s1.DP.x;
	dPLS1F.y = g.DP.y - s1.DP.y;
	dPLS1F.z = g.DP.z - s1.DP.z;
	ddUVLF.x = ddTLT[0][2];
	ddUVLF.y = ddTLT[1][2];
	ddUVLF.z = ddTLT[2][2];
	ddPLS1F.x = g.DDP.x - s1.DDP.x;
	ddPLS1F.y = g.DDP.y - s1.DDP.y;
	ddPLS1F.z = g.DDP.z - s1.DDP.z;

	VIV3 ( dUVLF, PLS1F, &w1 );
	VIV3 ( UVLF, dPLS1F, &da );
	da = da + w1;
	VIV3 ( ddUVLF, PLS1F, &w1 );
	VIV3 ( dUVLF, dPLS1F, &w2 );
	VIV3 ( UVLF, ddPLS1F, &dda );
	dda = dda + w1 + 2.0*w2;

	MPV3 ( dTLT, s2p, &dww );
	MPV3 ( ddTLT, s2p, &ddww );
	VIV3 ( dPLS1F, PLS1F,  &w1 );
	VIV3 ( dww, PLS1F,  &w2 );
	VIV3 ( ww, dPLS1F,  &w3 );
		
	db = 2.0*( w1 + w2 + w3 );
	VIV3 ( ddPLS1F, PLS1F,  &w1 );
	VIV3 ( dPLS1F, dPLS1F,  &w2 );
	VIV3 ( ddww, PLS1F,  &w3 );
	VIV3 ( dww, dPLS1F,  &w4 );
	VIV3 ( ww, ddPLS1F,  &w5 );
	ddb = 2.0*( w1 + w2 + w3 + 2.0*w4 + w5 );

	if( minv > 0 ) {
		the->v = -a + rootd;
        the->dv = -da + ( 2.0*da*a -  db )/( 2.0*rootd );
		the->ddv = -dda + ( 2.0*(2.0*dda*a + 2.0*da*da - ddb )*d - ( 2.0*da*a - db )*( 2.0*da*a - db ) )/( 4.0*d*rootd );
	}
	else {
		the->v = -a - rootd;
        the->dv = -da - ( 2.0*da*a -  db )/( 2.0*rootd );
		the->ddv = -dda - ( 2.0*(2.0*dda*a + 2.0*da*da - ddb )*d - ( 2.0*da*a - db )*( 2.0*da*a - db ) )/( 4.0*d*rootd );
	}

	p->P.x = the->v*UVLF.x + g.P.x;
	p->P.y = the->v*UVLF.y + g.P.y;
	p->P.z = the->v*UVLF.z + g.P.z;
	p->DP.x = the->dv*UVLF.x + the->v*dUVLF.x + g.DP.x;
	p->DP.y = the->dv*UVLF.y + the->v*dUVLF.y + g.DP.y;
	p->DP.z = the->dv*UVLF.z + the->v*dUVLF.z + g.DP.z;
	p->DDP.x = the->ddv*UVLF.x + 2.0*the->dv*dUVLF.x + the->v*ddUVLF.x + g.DDP.x;
	p->DDP.y = the->ddv*UVLF.y + 2.0*the->dv*dUVLF.y + the->v*ddUVLF.y + g.DDP.y;
	p->DDP.z = the->ddv*UVLF.z + 2.0*the->dv*dUVLF.z + the->v*ddUVLF.z + g.DDP.z;

	s2->P.x = p->P.x + ww.x;
	s2->P.y = p->P.y + ww.y;
	s2->P.z = p->P.z + ww.z;
	s2->DP.x = p->DP.x + dww.x;
	s2->DP.y = p->DP.y + dww.y;
	s2->DP.z = p->DP.z + dww.z;
	s2->DDP.x = p->DDP.x + ddww.x;
	s2->DDP.y = p->DDP.y + ddww.y;
	s2->DDP.z = p->DDP.z + ddww.z;

	return SUCCESS;
}

//-----------------------------------------------------------
//
//   ʒuxNg̏
//   
//   *j       : ʒuxNg
//
//-----------------------------------------------------------

void init_joint( POSITION *j )
{
	j->P.x = 0.0;
	j->P.y = 0.0;
	j->P.z = 0.0;

	j->DP.x = 0.0;
	j->DP.y = 0.0;
	j->DP.z = 0.0;

	j->DDP.x = 0.0;
	j->DDP.y = 0.0;
	j->DDP.z = 0.0;
}

//-----------------------------------------------------------
//
//   ʒuxNg̐̐ݒ
//   
//   x, y, z  : ʒuxNg̐
//   *j       : ʒuxNg
//
//-----------------------------------------------------------

void set_locate( double x, double y, double z, POINT *j )
{
	j->P.x = x;
	j->P.y = y;
	j->P.z = z;
}

//-----------------------------------------------------------
//
//   ʒuxNg̍xNg
//   
//   j1,j2@  : ʒuxNgƂ̔
//   *d       : xNgƂ̔
//
//-----------------------------------------------------------

void get_diff( POINT j1, POINT j2, POINT *d )
{
	d->P.x = j2.P.x - j1.P.x;
	d->P.y = j2.P.y - j1.P.y;
	d->P.z = j2.P.z - j1.P.z;
	d->DP.x = j2.DP.x - j1.DP.x;
	d->DP.y = j2.DP.y - j1.DP.y;
	d->DP.z = j2.DP.z - j1.DP.z;
	d->DDP.x = j2.DDP.x - j1.DDP.x;
	d->DDP.y = j2.DDP.y - j1.DDP.y;
	d->DDP.z = j2.DDP.z - j1.DDP.z;
}

//-----------------------------------------------------------
//
//   pψʍ\̂̐ݒ
//   
//   the, dthe, ddthe 
//            : pψʁCpxCpx
//   *angle   : pψʍ\
//
//-----------------------------------------------------------

void make_angle( double the, double dthe, double ddthe, ANGLE *angle )
{
	angle->the   = the;
	angle->dthe  = dthe;
	angle->ddthe = ddthe;
}
 
